###############################################
# 12 May 11
# 	Shared functions to analyze climate regression model output
#
###############################################

# install packages if necessary
packList <- c('fmsb', 'Hmisc', 'car')
for (pckg in packList) {
	if (!is.element(pckg, installed.packages()[,1])) {
		install.packages(pkgs=pckg, repos="http://cran.case.edu/", dependencies=TRUE)
	}
}

library('fmsb')		# for Nagelkerke pseudo-R2
library('Hmisc')	# for somers2()
library('car')		# for vif()


####################
# function for accessing polygon slot information, default is "ID"
####################
getPolySlot <- function(poly, sl="ID") {
	sapply(slot(poly, "polygons"), function(x) slot(x, sl)) 
}

####################
# function for to rescale variables by subtracting the mean & dividing by 1 or 2 sd
#	sdMult paramater determines the std. dev. multipler used to normalize the variable
#	as recommended by Andrew Gelman
#		http://www.stat.columbia.edu/~gelman/research/published/standardizing7.pdf
#	function modified from rescale() in the 'arm' package
#		install.packages(pkgs="arm", repos="http://cran.case.edu/", dependencies=TRUE)
#		library(arm)	# some conflicts w/ car (logit)
####################
myRescale <- function(x, sdMult) {
	# need to be careful about log-transformed variables -- elasticity interp
	if (!is.numeric(x)) {
        print("myRescale ERROR, variable is not numeric")
		return(FALSE)
    } 
	if (any(is.na(x))) {
        print("myRescale ERROR, variable contains NAs")
		return(FALSE)
	}
	if (sdMult != 1 && sdMult != 2) {
	    print("myRescale WARNING, sdMult should be 1 or 2")
	}
	(x - mean(x))/(sdMult * sd(x))
}

####################
# function that returns the elapsed time difference in a nice format
####################
elapsedTime <- function (startTime, endTime=NULL) {
	if (length(endTime) == 0)
		endTime <- proc.time()
	secDiff <- endTime[3]-startTime[3]
	if (secDiff <= 60+10) 
		return(paste(round(secDiff, digits=1), "seconds"))
	else if (secDiff <= 60*60+360)
		return(paste(round(secDiff/60, digits=1), "minutes"))
	else
		return(paste(round(secDiff/60/60, digits=1), "hours"))
}


####################
# calculate the area under the curve (AUC) for the given model
#	supports glmer models, otherwise
#		requires model$y to be defined or depVar to be provided
####################
calcAUC <- function (model, depVar = NULL) {
	# model <- lmObj; depVar=lmObj$model[, 1]
	fit <- fitted(model)		# faster than predict() 
								# but fitted() does not appear to support 'newdata'
#	fit <- predict(model, type="response")	# generally consistent across diff models
	fitBin <- fit
	fitBin[fit > 1] <- 1	# collapse to binary predicted variable
#	range(fit); range(fitBin)
	if (sum(fitBin < 0) > 0) {
		print("WARNING: negative values predicted, assigning them to 0")
		fitBin[fit < 0] <- 0
	}
	if (is(model, "merMod")) {	# sum(model@resp$y - model@frame$Events) = 0
		yBin <- (model@resp$y >= 1) * 1
	} else if (length(depVar) > 0) {
		yBin <- (depVar >= 1) * 1
	} else {
		yBin <- (model$y >= 1) * 1	# convert dependent var to binary
	}
	somers2(fitBin, yBin)[1]
#	somers2(plogis(fitBin), longData$EventsBin)[1]
	# plogis() only needed for predict values with type="link" (though return value is still the same)
}


####################
# function to calculate the predictive power of a given MLM model
#	1 apr 2014, add parameters to support manual multi-threading
#		- dropIs - indices from allPredVars to drop
#			first & last indices used to write unique file
#	requires valid 'outDir' object
####################
calcMLM_PredPower <- function(depVar, allPredVars, dataFr, funcForm, dropIs) {
	groupColName <- "Country"	# TODO: make this a parameter?
	#numVars <- length(allPredVars)
	numVars <- length(dropIs)
	AUCs <- rep(NA, numVars)
	names(AUCs) <- allPredVars[dropIs]
	filename <- paste(outDir, "calcMLM_AUCs_", dropIs[1],"_", dropIs[numVars],".csv", sep="")
	print(paste("Will write partial results to:", filename))
	startTime <- proc.time()
	for (i in dropIs) {	# 1:numVars
		modelStartTime <- proc.time()
		print(paste("running MLM after removing", allPredVars[i], ". . .")); flush.console()
		currIndVars <- allPredVars[-i]
		mlmText <- paste(depVar, " ~ ", paste(currIndVars, "+ ", collapse=""), sep="")
		mlmFormula <- as.formula(paste(mlmText,"(1 | ",groupColName,")", sep=""))
		if (funcForm == "linear") {
			lmerObj <- lmer(mlmFormula, data=dataFr)	# consider REML = FALSE?
			modelName <- paste("lmer_no", allPredVars[i], sep="")
		} else if (funcForm == "poisson") {
			lmerObj <- glmer(mlmFormula, data=dataFr, family=poisson)
			modelName <- paste("glmer_pois_no", allPredVars[i], sep="")
		}
		else
			print("ERROR: invalid function form")
		# write results to file to check stability of parameters
		writeLMER(lmerObj, paste(modelName, ".csv", sep=""), groupColName)
		AUCs[i] <- calcAUC(lmerObj)
		# write results to file after each glmer() & calcAUC() call
		write.csv(AUCs, file=filename)	# TODO: don't need to write this anymore
		print(paste("     completed in: ", elapsedTime(modelStartTime), sep=""))
	}
	print(paste("calcMLM_PredPower() execution time: ", elapsedTime(startTime), sep=""))
	AUCs
}

print("Loaded shared climate model functions")

