###################################
# 26 Sep 2012
# 	shared multilevel model functions 
#	some package comparison info for mixed/multilevel modeling
#		http://glmm.wikidot.com/pkg-comparison
#	multilevel model notation:
#		http://glmm.wikidot.com/faq#toc26
#
###################################

# install packages if necessary
packList <- c('lme4')
for (pckg in packList) {
	if (!is.element(pckg, installed.packages()[,1])) {
		install.packages(pkgs=pckg, repos="http://cran.case.edu/", dependencies=TRUE)
	}
}
library(lme4)


###################################
# function to write model to disk for LMER (lme4) models
#	default is 'Country' as the group level parameter
###################################
writeLMER <- function (lme4model, fname, groupColName="Country") {
#	lme4model <- lmerObj; lme4model <- glmer_logit; groupColName <- "ccode"
#	str(lme4model)
	cfs <- coef(summary(lme4model))	#[,-2] # keep the standard errors to compare with randEffects 
	# 25 Nov 2013, "postVar is deprecated: please use condVar instead"
	#	http://tolstoy.newcastle.edu.au/R/help/06/08/31936.html
	u0 <- lme4::ranef(lme4model, condVar=TRUE)	# str(u0)
	# only calculated SEs for 'simple' random effects models
	# 	since condVar=TRUE is not yet implemented for multiple terms
	if (dim(attr(u0[[groupColName]], "postVar"))[1] == 1) {
		#u0se <- sqrt(attr(u0[[1]], "postVar")[1, , ])	# but attribute still called "postVar"!
		#u0se <- sqrt(attr(u0$Grid, "postVar")[1, , ])
		#u0se <- sqrt(attr(u0$Country, "postVar")[1, , ])
		u0se <- sqrt(attr(u0[[groupColName]], "postVar")[1, , ])
		# calculate z-value for random effects
		if (ncol(cfs) == 3) {
			randEffects <- cbind(u0[[groupColName]], u0se, u0[[groupColName]]/u0se)
			colnames(randEffects) <- colnames(cfs)	# required for rbind() to work
			# insert actual column names as a row
			randEffects <- rbind(c(paste(groupColName,"Intercept",sep=""), "StdError", "z value"), randEffects)
		} else if (ncol(cfs) == 4) {
			randEffects <- cbind(u0[[groupColName]], u0se, u0[[groupColName]]/u0se, NA)
			colnames(randEffects) <- colnames(cfs)	# required for rbind() to work
			randEffects <- rbind(c(paste(groupColName,"Intercept",sep=""), "StdError", "z value", ""), randEffects)
		} else
			print("ERROR: unexpected number of columns in coef obj") 
	} else {
		randEffects <- cbind(u0[[groupColName]], NA)
		origColNames <- colnames(randEffects)
		colnames(randEffects) <- colnames(cfs)	# required for rbind() to work
		randEffects <- rbind(c(paste(groupColName,"Intercept",sep=""), paste(origColNames[2],"Slope",sep=""), origColNames[3]), randEffects)
	}
	groupVarName <- paste(groupColName, "Variance", sep="")
	groupVPCname <- paste(groupColName, "VarPartitionCoef", sep="")
	diagnosticVars <- c("logLik", "AIC", groupVarName, "Lev1Variance", groupVPCname, 
						"AUC", "N", "SumEvents", "Residuals(s.d.)",
						"GridVariance", "GridVarPartitionCoef")
	outData <- rbind(cfs, randEffects, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA)
	rownames(outData) <- c(rownames(cfs), rownames(randEffects), diagnosticVars)
	outData["logLik", 1] <- logLik(lme4model)
	outData["AIC", 1] <- AIC(lme4model)
	lmeVarCorr <- VarCorr(lme4model)
	groupVar <- lmeVarCorr[[groupColName]][1] # extract the group (level 2) variance
		# 11 apr 2014, made a brief effort to calculate the SE for groupVar, but gave up -- can use bootstrap if necessary
	outData[groupVarName, 1] <- groupVar
	# calculate variance partition coeff
	if (is(lme4model)[1] == "lmerMod") {
		lev1Var <- attr(lmeVarCorr, "sc")^2	# extract the remaining (level 1) var.
	} else if (is(lme4model)[1] == "glmerMod") {
		print("WARNING: assuming constant level 1 var = 3.29 for logit glmer")
		print(paste("calculated value is:", attr(lmeVarCorr, "sc")^2))
		lev1Var <- 3.29
		# other (level 1) variance = 3.29 for logit link (pi^2)/3)
		# C:\Users\witmer\Documents\CU\ClimateConflict\SubSahAfrica\Analysis\RegressionModels\Multilevel_RE\R-7_binMultilevel.pdf
		# probably should not be used for log links! http://stats.stackexchange.com/questions/25371/variance-explained-by-random-effects-using-lme4
	} else
		print("WARNING: unrecognized model object to writeLMER()")
	outData["Lev1Variance", 1] <- lev1Var
	VPC <- round(groupVar/(groupVar + lev1Var)*100, digits=2)
	outData[groupVPCname, 1] <- VPC
	outData["AUC", 1] <- calcAUC(lme4model)
	#outData["AUC", 1] <- calcAUC(lme4model, lme4model@resp$y)
	outData["N", 1] <- length(lme4model@resp$y)
	outData["SumEvents", 1] <- sum(lme4model@resp$y)
	outData["Residuals(s.d.)", 1] <- sd(residuals(lme4model))
	gridVar <- lmeVarCorr$Grid[1]	# retain, even if empty
	if (length(gridVar) != 0) {
		outData["GridVariance", 1] <- gridVar
		# TODO: check this & groupVPC calculation when levels > 2
		outData["GridVarPartitionCoef", 1] <- round(gridVar/(gridVar+3.29)*100, digits=2)
	}
	write.csv(outData, file=paste(outDir, fname, sep=""))
}

###################################
# plot residuals with error bars by country for LMER (lme4) models
#	95% CI calculated using the posterior variances of the random effects (sometimes referred to as conditional modes/variances) http://figuraleffect.wordpress.com/2011/07/01/computing-standard-errors-ses-and-confindence-intervals-for-blups-in-lmerlme4/
#	these plots of the conditional modes are known as caterpillar plots
#		http://www.ats.ucla.edu/stat/r/dae/melogit.htm
#	default is 'Country' as the group level parameter
###################################
residPlotLMER <- function(lme4model, filePrefix, groupColName="Country", groupLookup=NULL) {
#	lme4model <- lmerObj; groupColName <- "ccode"; groupLookup <- countryLookup
	u0 <- lme4::ranef(lme4model, condVar=TRUE)
	#u0se <- sqrt(attr(u0[[1]], "postVar")[1, , ])	# but attribute still called "postVar"!
	#u0tab <- cbind(countryID, u0[[1]], u0se)
	#countryID <- rownames(u0[[1]])
	#u0se <- sqrt(attr(u0$Country, "postVar")[1, , ])
	u0se <- sqrt(attr(u0[[groupColName]], "postVar")[1, , ])
	groupID <- rownames(u0[[groupColName]])
	u0tab <- cbind(groupID, u0[[groupColName]], u0se)
	colnames(u0tab) <- c("groupID", "u0", "u0se")
	u0tab <- u0tab[order(u0tab$u0), ]	# sort
	u0tab <- cbind(u0tab, c(1:dim(u0tab)[1]))
	colnames(u0tab)[4] <- "u0rank"
	u0tab <- u0tab[order(u0tab$groupID), ]
	lWidth <- 1.4
	barCol <- "gray50"
	lmeVarCorr <- VarCorr(lme4model)
	groupVar <- lmeVarCorr[[groupColName]][1] # extract the group (level 2) variance
	# calculate variance partition coeff
	if (is(lme4model)[1] == "lmerMod"){
		lev1Var <- attr(lmeVarCorr, "sc")^2	# extract the remaining (level 1) var.
	} else if (is(lme4model)[1] == "glmerMod") {
		lev1Var <- 3.29
		print("WARNING: assuming constant level 1 var = 3.29 for logit glmer")
		# other (level 1) variance = 3.29 for logit link (pi^2)/3)
		# C:\Users\witmer\Documents\CU\ClimateConflict\SubSahAfrica\Analysis\RegressionModels\Multilevel_RE\R-7_binMultilevel.pdf
	} else
		print("WARNING: unrecognized model object to writeLMER()")
	VPC <- round(groupVar/(groupVar + lev1Var)*100, digits=1)
	#browser()
	#range(u0tab$u0)
	CIbuff <- mean(1.96*u0tab$u0se)/4
	yMin <- min(-1, min(u0tab$u0) - CIbuff)	#- 0.3
	yMax <- max(1, max(u0tab$u0) + CIbuff)	#+ 0.2
	ySpan <- yMax - yMin
	pdf(paste(outDir, filePrefix, ".pdf", sep=""), width=10, height=7)
		plot(u0tab$u0rank, u0tab$u0, xlim=c(1,nrow(u0tab)), ylim=c(yMin,yMax), type = "n", xlab = "u rank", ylab = paste(groupColName, "level residuals (95% confidence intervals)"), main=paste("Multilevel Model", groupColName, "Residuals"), axes=FALSE)	# type = n creates axis for the plot but prevents data from being plotted
		axis(2)	# add default y-axis
		axis(1, at=1:nrow(u0tab))
		segments(1, 0, nrow(u0tab), 0, lty=2, col="red", lwd=lWidth)	# 0 line
		if (sum(u0tab$u0se) > 0.001) {
			# add to the plot the 95% confidence intervals
			tmp<-sapply(1:nrow(u0tab), function(i) arrows(u0tab$u0rank[i],u0tab$u0[i]-1.96*u0tab$u0se[i],u0tab$u0rank[i],u0tab$u0[i]+1.96*u0tab$u0se[i], angle=90,code=3,length=.04,col=barCol,lwd=lWidth) )
		}
		points(u0tab$u0rank, u0tab$u0, pch=20, col=barCol)
			# note: low ranks ==> overpredicted; high ==> underpredicted
		#brdrCntryI <- c(2,7,8,11,14)
		lblFsize <- 0.6
		if (length(groupLookup) > 0)
			groupLabels <- groupLookup[match(countryLookup[[groupColName]], u0tab$groupID), 1]
		else
			groupLabels <- u0tab$groupID
		text(u0tab$u0rank-0.1, u0tab$u0 + CIbuff, groupLabels, srt=0, cex=lblFsize, adj=c(0,0), srt=45, col="blue")
		#text(u0tab$u0rank[brdrCntryI]-0.1, u0tab$u0[brdrCntryI] + 1.96*u0tab$u0se[brdrCntryI]+0.05, u0tab$groupID[brdrCntryI], srt=0, cex=lblFsize, adj=c(0,0))
		#text(u0tab$u0rank[-brdrCntryI]-0.1, u0tab$u0[-brdrCntryI] + 1.96*u0tab$u0se[-brdrCntryI]+0.05, u0tab$groupID[-brdrCntryI], srt=0, cex=lblFsize, adj=c(0,0), col="blue")
		#text(11, -1.5, adj=0, cex=lblFsize, "Core country", col="blue")
		#text(11, -1.7, adj=0, cex=lblFsize, "Border country")
		text(1, yMax-(ySpan*0.1), adj=0, cex=1, paste(groupColName, " variance partition coefficient = ", VPC, "%", sep=""))
	dev.off()
#	setwd(outDir); shell(paste("convert -trim -bordercolor White -border 20 -density 200 ", filePrefix, ".pdf ", filePrefix, ".jpg", sep=""))
}

####################
# calculate means & then demean according to the groupColName (e.g. Country or Grid) 
#	in preparation for Bell & Jones style model
#	returns modified dataFr
####################
meanDemean <- function(dataFr, vars, groupColName) {
	startTime <- proc.time()
	meanVars <- paste(vars, "_mean", sep="")
	demeanVars <- paste(vars, "_demean", sep="")
	for (j in 1:length(vars)) {
		dataFr[[meanVars[j]]] <- NA
	}
	groupNames <- unique(dataFr[[groupColName]])
	# could try to re-write this w/ aggregate() to remove one of the loops, but this works...
	for (i in 1:length(groupNames)) {
		grp <- groupNames[i]
		subdata <- dataFr[dataFr[[groupColName]]==grp,]
		for (j in 1:length(vars)) {
			dataFr[[meanVars[j]]] <- ifelse(dataFr[[groupColName]]==grp, mean(subdata[[vars[j]]], na.rm=T), dataFr[[meanVars[j]]])
		}
	}
	for (j in 1:length(vars)) {
		dataFr[[demeanVars[j]]] <- dataFr[[vars[j]]] - dataFr[[meanVars[j]]]
	}
	print(paste("meanDemean() execution time: ", elapsedTime(startTime), sep=""))
	dataFr
}


####################
# center list of vars using grand mean (cgm)
#	in preparation for Bell & Jones style model
#	returns modified dataFr
####################
centerVars <- function(dataFr, vars) {
	centerVars <- paste(vars, "_cgm", sep="")
	for (j in 1:length(vars)) {
		dataFr[[centerVars[j]]] <- dataFr[[vars[j]]] - mean(dataFr[[vars[j]]])
	}
	dataFr
}

####################
# helper function that appends "_mean" and "_demean" to 'vars'
#	return list is interspersed
####################
appendMeanDemean <- function(vars) {
	meanVars <- paste(vars, "_mean", sep="")
	demeanVars <- paste(vars, "_demean", sep="")
	# intersperse mean & demean vars
	paste(sapply(1:length(vars), function(i) c(meanVars[i], demeanVars[i])))
}

###################################
# function to estimate Bell & Jones style multilevel/RE model
#	uses lmer
#	requires _mean & _demean predVars to have already been calculated
#	writes output to disk
#	tradVars are added as is
#	returns lmerObj
###################################
BellJonesLMER <- function(depVar, predVars, dataFr, modelName, groupColName, groupLookup, funcForm="ols", tradVars=NULL, writeFiles=TRUE) {
	startTime <- proc.time()
	if (length(predVars) > 0) {
		meanDemeanVars <- appendMeanDemean(predVars)
		if (length(tradVars) == 0)
			allPredVars <- meanDemeanVars
		else
			allPredVars <- c(meanDemeanVars, tradVars)
		lmText <- paste(depVar, " ~ ", paste(allPredVars, "+ ", collapse=""), sep="")
	} else {
		print("Estimating null multilevel model with no predictors")
		lmText <- paste(depVar, " ~ ", sep="")
	}
	lmFormula <- as.formula(paste(lmText,"(1 | ",groupColName,")", sep=""))
	if (funcForm == "ols" )
		lmerObj <- lmer(lmFormula, data=dataFr)	# consider REML = FALSE?
	else if (funcForm == "logit")
		lmerObj <- glmer(lmFormula, data=dataFr, family=binomial(logit))
	else
		print("ERROR: invalid function form")
	#summary(lmerObj)	#fixef(lmerObj)
	if (writeFiles) {
		writeLMER(lmerObj, paste(modelName, ".csv", sep=""), groupColName=groupColName)
		residPlotLMER(lmerObj, paste(modelName, groupColName, "Residuals", sep=""), groupColName=groupColName, groupLookup=groupLookup)
	}
	print(paste("BellJonesLMER() execution time: ", elapsedTime(startTime), sep=""))
	lmerObj
}

###################################
# run given MLM formula & functional form on the full dataset
#	requires: outDir, writeLMER(), residPlotLMER()
#	returns: lmerObj
###################################
fullLMER <- function(dataFr, mlmForm, funcForm) {
	startTime <- proc.time()
	groupColName <- "Country"
	currTime <- Sys.time()
	if (funcForm == "linear") {
		lmerObj <- try(lmer(mlmForm, data=dataFr))	# consider REML = FALSE?
		modelPrefix <- "lmer_full_"
	} else if (funcForm == "poisson") {
		# encapsulate call in try() to handle convegence failures
		lmerObj <- try(glmer(mlmForm, data=dataFr, family=poisson))
		modelPrefix <- "glmer_pois_full_"
	}
	if (class(lmerObj) == "try-error") {
		print("WARNING: try-error encountered!")
		return(NA)
	}
	modelName <- paste(modelPrefix,format(currTime, "%d%b%y_%H_%M_%S"), sep="")
	# use system time to write results to unique file in case system crashes!
	writeLMER(lmerObj, paste(modelName,".csv", sep=""), groupColName = groupColName)
	residPlotLMER(lmerObj, paste(modelName,"CountryResiduals", sep=""), groupColName = groupColName)
	print(paste("    fullLMER call complete in", elapsedTime(startTime))); flush.console()
	lmerObj
}

###################################
# recursive function that resamples data by cluster/group level
# source:
#	http://biostat.mc.vanderbilt.edu/wiki/Main/HowToBootstrapCorrelatedData
# optimal usage: improve memory usage by only passing it the minimum data:
#		a unique id and the cluster fields
###################################
resampleMLM <- function(dat, cluster, replace) { # replace <- c(T,T,T)
  # exit early for trivial data
  if(nrow(dat) == 1 || all(replace==FALSE)) {
	  print(paste("no sample taken, nrow(dat) =", nrow(dat))); flush.console()
      return(dat)
	}
  #print(paste("resampling for level:", cluster[1])); flush.console()
  
  # sample the clustering factor
  cls <- sample(unique(dat[[cluster[1]]]), replace=replace[1])
  
  # subset on the sampled clustering factors
  #		returns a list of length equal to the number of primary clusters
  sub <- lapply(cls, function(b) subset(dat, dat[[cluster[1]]]==b))
  
  # sample lower levels of hierarchy (if any)
  if(length(cluster) > 1)
    sub <- lapply(sub, resampleMLM, cluster=cluster[-1], replace=replace[-1])
	# resampleMLM(sub[[1]], cluster=cluster[-1], replace=replace[-1])
  
  # join and return samples
  do.call(rbind, sub)
}

###################################
# run given MLM formula & functional form on the specified sampled indices
#	requires: outDir, writeLMER(), residPlotLMER()
#	returns: model coeff. estimates
###################################
sampleLMER <- function(dataFr, mlmForm, funcForm, threadID, indices) {
	startTime <- proc.time()
	print(paste("thread", threadID, "first 5 indices:"))
	print(indices[1:5]); flush.console()
	groupColName <- "Country"
	currTime <- Sys.time()
	dataSample <- dataFr[indices, ]
	if (funcForm == "linear") {
		lmerObj <- try(lmer(mlmForm, data=dataSample))	# consider REML = FALSE?
		modelPrefix <- "lmer_boot_"
	} else if (funcForm == "poisson") {
		# encapsulate call in try() to handle convegence failures
		lmerObj <- try(glmer(mlmForm, data=dataSample, family=poisson))
		modelPrefix <- "glmer_pois_boot_"
	}
	if (class(lmerObj) == "try-error") {
		print("WARNING: try-error encountered!")
		return(NA)
	}
	modelName <- paste(modelPrefix,threadID,"_",format(currTime, "%d%b%y_%H_%M_%S"), sep="")
	# use system time to write results to unique file in case system crashes!
	writeLMER(lmerObj, paste(modelName,".csv", sep=""), groupColName = groupColName)
	#residPlotLMER(lmerObj, paste(modelName,"CountryResiduals", sep=""), groupColName = groupColName)
	print(paste("    sampleLMER call complete in", elapsedTime(startTime))); flush.console()
	# TODO: also return N & % non-zero grid cells in revised sample?
	# TODO: make sure length of estimates == number of coeffs!?
	coef(summary(lmerObj))[, "Estimate"]
}

# could probably combine sampleLMER() and bootLoop() into 1, but they are separate due to old use of 'boot' package

###################################
# function that resamples multilevel data and estimates coefficients
#	designed to be called from lapply style statement
###################################
bootLoop <- function(x) {
	startTime <- proc.time()
	currSample <- resampleMLM(pastDataSmall, datLevels, rep(T, length(datLevels)))
	#currSample <- resampleMLM(pastDataSmall, datLevels, c(T,T,T))
	#print(paste("resampleMLM() execution time: ", elapsedTime(startTime), sep="")); flush.console() # 1-1.5 mins
	sampleLMER(data=pastData, mlmForm=mlmFormula, funcForm=funcForm, threadID=paste(threadNum, x, sep="_"), indices=currSample$Id)
}


print("Loaded shared multilevel functions")
